/*
								+---------------------------------+
								|                                 |
								|      ***   OpenGL2.0   ***      |
								|                                 |
								|  Copyright   -tHE SWINe- 2006  |
								|                                 |
								|         RenderBuffer2.h         |
								|                                 |
								+---------------------------------+
*/

/*
 *	passed code revision at 10-05-2006
 *
 *	integers, containing just true and false were replaced with bool
 *	fixed some design issues to enable binding depth and stencil output texture along with
 *	color texture
 *	WGL_ARB_pixel_buffer support was deprecated
 *	files were renamed to RenderBuffer2.cpp and RenderBuffer2.h respectively
 *
 *	03-06-2007
 *
 *	removed __declspec(dllexport)
 *
 */

#ifndef __RENDER_BUFFER2_INCLUDED
#define __RENDER_BUFFER2_INCLUDED

/*
 *								=== CGLRenderBuffer ===
 */

class /*__declspec(dllexport)*/ CGLRenderBuffer {
protected:
	CGLState *m_p_state;
	int m_n_width;
	int m_n_height;

	int m_n_draw_buffer_num;
	GLenum m_n_color_internal_format;

	bool m_b_depth_buffer;
	GLenum m_n_depth_internal_format;

	bool m_b_stencil_buffer;
	GLenum m_n_stencil_internal_format;

	bool m_b_color_texture_target;
	bool m_b_depth_texture_target;
	bool m_b_stencil_texture_target;

	GLuint m_p_bound_color_texture[16];
	int m_n_bound_color_texture_num; // number of textures, being currently bound
	GLuint m_n_bound_depth_texture;
	GLuint m_n_bound_stencil_texture; // current bound texture or 0

	bool m_b_status;
	bool m_b_active;

public:
	/*
	 *	CGLRenderBuffer(int n_width, int n_height,
	 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
	 *		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
	 *		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
	 *		- constructor; creates the render buffer (OpenGL must be initialized)
	 *		- n_width and n_height are dimensions of the output buffer
	 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
	 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
	 *		- b_color_texture_target enables binding color texture(s) as render target(s)
	 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
	 *		- n_color_internal_format is OpenGL internal texture format (such as GL_RGB8, ...)
	 *		- b_depth_buffer enables depth output
	 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
	 *		  (otherwise depth fragments are written into internal buffer)
	 *		- n_depth_internal_format is depth buffer pixel format (one of GL_DEPTH_COMPONENT16,
	 *		  GL_DEPTH_COMPONENT24 or GL_DEPTH_COMPONENT32)
	 *		- b_stencil_buffer enables stencil output
	 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
	 *		  (otherwise stencil fragments are written into internal buffer)
	 *		- n_stencil_internal_format is stencil buffer pixel format (one of
	 *		  GL_STENCIL_INDEX1_EXT, GL_STENCIL_INDEX4_EXT, GL_STENCIL_INDEX8_EXT or
	 *		  GL_STENCIL_INDEX16_EXT)
	 */
	CGLRenderBuffer(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format);

	/*
	 *	CGLRenderBuffer(int n_width, int n_height,
	 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
	 *		int n_color_bit_num, bool b_float_color,
	 *		bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
	 *		bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num)
	 *		- constructor; creates the render buffer (OpenGL must be initialized)
	 *		- n_width and n_height are dimensions of the output buffer
	 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
	 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
	 *		- b_color_texture_target enables binding color texture(s) as render target(s)
	 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
	 *		- n_color_format is OpenGL texture format (one of GL_RGB, GL_RGBA, GL_INTENSITY,
	 *		  GL_LUMINANCE, GL_ALPHA, GL_LUMINANCE_ALPHA)
	 *		- n_color_bit_num is color bit depth (4, 8 or 16 for integer, 16 or 32 for float formats)
	 *		- b_float_color decides between integer (false) or float (true) formats
	 *		- b_depth_buffer enables depth output
	 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
	 *		  (otherwise depth fragments are written into internal buffer)
	 *		- n_depth_bit_num is depth buffer pixel width (16, 24 or 32)
	 *		- b_stencil_buffer enables stencil output
	 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
	 *		  (otherwise stencil fragments are written into internal buffer)
	 *		- n_stencil_bit_num is stencil buffer pixel width (1, 4, 8 or 16)
	 */
	CGLRenderBuffer(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
		int n_color_bit_num, bool b_float_color,
		bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
		bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num);

	/*
	 *	virtual CGLRenderBuffer::~CGLRenderBuffer()
	 *		- default destructor
	 *		- takes care of cleaning up
	 */
	virtual ~CGLRenderBuffer() {}

	/*
	 *	static bool CGLRenderBuffer::b_Supported()
	 *		- returns true in case the render buffer class have necessary hardware support
	 */
	static bool b_Supported();

	/*
	 *	static int CGLRenderBuffer::n_Max_DrawBuffer_Num()
	 *		- returns maximal number of OpenGL draw buffers (1 where not supported, otherwise
	 *		  1, 2, 4, up to 16 in future)
	 */
	static int n_Max_DrawBuffer_Num();

	/*
	 *	virtual bool CGLRenderBuffer::b_Status() const
	 *		- returns true in case render buffer, specified by constructor parameters
	 *		  was successfuly created and is ready to use
	 */
	virtual bool b_Status() const;

	/*
	 *	int CGLRenderBuffer::n_Width() const
	 *		- returns width of the render buffer
	 */
	int n_Width() const;

	/*
	 *	int CGLRenderBuffer::n_Height() const
	 *		- returns height of the render buffer
	 */
	int n_Height() const;

	/*
	 *	virtual bool CGLRenderBuffer::HandleModeSwitch()
	 *		- see if the render-buffer was lost during display mode switch
	 *		- returns true in case render-buffer is ok or false in case render buffer
	 *		  was lost (creates a new buffer in latter case, the result can be obtained
	 *		  by call to b_Status())
	 */
	virtual bool HandleModeSwitch() = 0;

	/*
	 *	virtual bool CGLRenderBuffer::Bind()
	 *		- binds the render buffer as OpenGL output buffer
	 *		- returns true on success or false on failure
	 */
	virtual bool Bind() = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type, int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type, int n_index = 0) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(const CGLTexture &r_texture,
	 *		int n_z_offset, Glenum int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_z_offset is index of 2D slice to be rendered to
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_3D(const CGLTexture &r_texture, int n_z_offset, int n_index = 0) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for depth output
	 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindDepthTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for stencil output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindStencilTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type, int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_2D(GLuint n_texture, GLenum n_texture_type, int n_index = 0) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(GLuint n_texture,
	 *		int n_z_offset, Glenum int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_z_offset is index of 2D slice to be rendered to
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_3D(GLuint n_texture, int n_z_offset, int n_index = 0) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for depth output
	 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindDepthTexture_2D(GLuint n_texture, GLenum n_texture_type) = 0;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for stencil output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindStencilTexture_2D(GLuint n_texture, GLenum n_texture_type) = 0;

	/*
	 *	virtual CGLState *CGLRenderBuffer::p_State()
	 *		- returns render-buffer OpenGL state guard (should be identic to
	 *		  global state guard for FBO, different for PBO)
	 *		- returns 0 in case render buffer is not initialized / ready
	 */
	virtual CGLState *p_State() = 0;

	/*
	 *	virtual bool CGLRenderBuffer::CopyTextureData(bool b_release_textures)
	 *		- copies texture data to textures it wasn't possible to directly render to
	 *		- in case b_release_textures is true, all currently bound textures are released
	 *		- returns true on success or false on failure
	 */
	virtual bool CopyTextureData(bool b_release_textures) = 0;

	/*
	 *	virtual bool CGLRenderBuffer::b_DirectRenderToTexture() const
	 *		- returns true in case direct rendering to all textures, being bound is supported,
	 *		  otherwise false
	 */
	virtual bool b_DirectRenderToTexture() const = 0;

	/*
	 *	virtual bool CGLRenderBuffer::Release()
	 *		- releases the render buffer from OpenGL output buffer binding
	 *		- returns true on success or false on failure
	 */
	virtual bool Release() = 0;

protected:
	/*
	 *	virtual bool CGLRenderBuffer::Create()
	 *		- take member variables, set by constructor and attempt to create the render buffer
	 *		- returns true on success or false on failure
	 *		- m_b_status is set to return value of this function afterwards
	 */
	virtual bool Create() = 0;
	static GLenum n_InternalColorFormat(GLenum n_color_format, int n_bit_num, bool b_float);
};

/*
 *								=== ~CGLRenderBuffer ===
 */

/*
 *	(PBuffers are not supported at this time - todo?)
 */

/*
 *								=== CGLRenderBuffer_FBO ===
 */

class /*__declspec(dllexport)*/ CGLRenderBuffer_FBO : public CGLRenderBuffer {
protected:
	GLuint m_n_framebuffer;
	GLuint m_p_color_rb[16];
	GLuint m_n_depth_rb;
	GLuint m_n_stencil_rb; // FBO objects

	GLuint m_p_draw_buffers[16]; // draw buffers state, saved by Bind() and
	// restored by Release() in case more than one draw-buffer is used

public:
	/*
	 *	CGLRenderBuffer_FBO(int n_width, int n_height,
	 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
	 *		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
	 *		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format)
	 *		- constructor; creates the render buffer (OpenGL must be initialized)
	 *		- n_width and n_height are dimensions of the output buffer
	 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
	 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
	 *		- b_color_texture_target enables binding color texture(s) as render target(s)
	 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
	 *		- n_color_internal_format is OpenGL internal texture format (such as GL_RGB8, ...)
	 *		- b_depth_buffer enables depth output
	 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
	 *		  (otherwise depth fragments are written into internal buffer)
	 *		- n_depth_internal_format is depth buffer pixel format (one of GL_DEPTH_COMPONENT16,
	 *		  GL_DEPTH_COMPONENT24 or GL_DEPTH_COMPONENT32)
	 *		- b_stencil_buffer enables stencil output
	 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
	 *		  (otherwise stencil fragments are written into internal buffer)
	 *		- n_stencil_internal_format is stencil buffer pixel format (one of
	 *		  GL_STENCIL_INDEX1_EXT, GL_STENCIL_INDEX4_EXT, GL_STENCIL_INDEX8_EXT or
	 *		  GL_STENCIL_INDEX16_EXT)
	 */
	CGLRenderBuffer_FBO(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_internal_format,
		bool b_depth_buffer, bool b_depth_texture_target, GLenum n_depth_internal_format,
		bool b_stencil_buffer, bool b_stencil_texture_target, GLenum n_stencil_internal_format);

	/*
	 *	CGLRenderBuffer_FBO(int n_width, int n_height,
	 *		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
	 *		int n_color_bit_num, bool b_float_color,
	 *		bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
	 *		bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num)
	 *		- constructor; creates the render buffer (OpenGL must be initialized)
	 *		- n_width and n_height are dimensions of the output buffer
	 *		- n_color_buffer_num is number of draw buffers (0 / 1 / 2 / 4 on geforce, 8 on quadro,
	 *		  up to 16 in future; 0 means no color output is generated - depth map rendering)
	 *		- b_color_texture_target enables binding color texture(s) as render target(s)
	 *		  otherwise color is rendered into internal buffer (glReadPixels() / glCopyTexImage*())
	 *		- n_color_format is OpenGL texture format (one of GL_RGB, GL_RGBA, GL_INTENSITY,
	 *		  GL_LUMINANCE, GL_ALPHA, GL_LUMINANCE_ALPHA)
	 *		- n_color_bit_num is color bit depth (4, 8 or 16 for integer, 16 or 32 for float formats)
	 *		- b_float_color decides between integer (false) or float (true) formats
	 *		- b_depth_buffer enables depth output
	 *		- b_depth_texture_target enables writing depth fragments into bound depth texture
	 *		  (otherwise depth fragments are written into internal buffer)
	 *		- n_depth_bit_num is depth buffer pixel width (16, 24 or 32)
	 *		- b_stencil_buffer enables stencil output
	 *		- b_stencil_texture_target enables writing stencil fragments into bound stencil texture
	 *		  (otherwise stencil fragments are written into internal buffer)
	 *		- n_stencil_bit_num is stencil buffer pixel width (1, 4, 8 or 16)
	 */
	CGLRenderBuffer_FBO(int n_width, int n_height,
		int n_color_buffer_num, bool b_color_texture_target, GLenum n_color_format,
		int n_color_bit_num, bool b_float_color,
		bool b_depth_buffer, bool b_depth_texture_target, int n_depth_bit_num,
		bool b_stencil_buffer, bool b_stencil_texture_target, int n_stencil_bit_num);

	/*
	 *	virtual CGLRenderBuffer_FBO::~CGLRenderBuffer_FBO()
	 *		- default destructor
	 *		- takes care of cleaning up
	 */
	virtual ~CGLRenderBuffer_FBO();

	/*
	 *	static bool CGLRenderBuffer_FBO::b_Supported() const
	 *		- returns true in case the render buffer class have necessary hardware support
	 */
	static bool b_Supported();

	/*
	 *	static int CGLRenderBuffer_FBO::n_Max_DrawBuffer_Num() const
	 *		- returns maximal number of OpenGL draw buffers (1 where not supported, otherwise
	 *		  1, 2, 4, up to 16 in future)
	 */
	static int n_Max_DrawBuffer_Num();

	/*
	 *	virtual bool CGLRenderBuffer_FBO::HandleModeSwitch()
	 *		- see if the render-buffer was lost during display mode switch
	 *		- returns true in case render-buffer is ok or false in case render buffer
	 *		  was lost (creates a new buffer in latter case, the result can be obtained
	 *		  by call to b_Status())
	 */
	virtual bool HandleModeSwitch();

	/*
	 *	virtual bool CGLRenderBuffer_FBO::Bind()
	 *		- binds the render buffer as OpenGL output buffer
	 *		- returns true on success or false on failure
	 */
	virtual bool Bind();

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type, int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type, int n_index = 0);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(const CGLTexture &r_texture,
	 *		int n_z_offset, Glenum int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_z_offset is index of 2D slice to be rendered to
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_3D(const CGLTexture &r_texture, int n_z_offset, int n_index = 0);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for depth output
	 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindDepthTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(const CGLTexture &r_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for stencil output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindStencilTexture_2D(const CGLTexture &r_texture, GLenum n_texture_type);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type, int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_2D(GLuint n_texture, GLenum n_texture_type, int n_index = 0);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindColorTexture_3D(GLuint n_texture,
	 *		int n_z_offset, Glenum int n_index = 0)
	 *		- binds OpenGL texture n_texture as target for color output
	 *		- n_z_offset is index of 2D slice to be rendered to
	 *		- n_index is index of one of draw buffers texture is to be bound to
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindColorTexture_3D(GLuint n_texture, int n_z_offset, int n_index = 0);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindDepthTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for depth output
	 *		- n_texture_type is either GL_TEXTURE_2D or GL_TEXTURE_RECTANGLE
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindDepthTexture_2D(GLuint n_texture, GLenum n_texture_type);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::BindStencilTexture_2D(GLuint n_texture,
	 *		GLenum n_texture_type)
	 *		- binds OpenGL texture n_texture as target for stencil output
	 *		- n_texture_type is either GL_TEXTURE_2D, GL_TEXTURE_RECTANGLE,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_X, GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	 *		  GL_TEXTURE_CUBE_MAP_POSITIVE_Z, GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	 *		  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y, or GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
	 *		- binding texture with id 0 releases previously bound texture
	 *		- returns true on success or false on failure
	 */
	virtual bool BindStencilTexture_2D(GLuint n_texture, GLenum n_texture_type);

	/*
	 *	virtual CGLState *CGLRenderBuffer_FBO::p_State()
	 *		- returns render-buffer OpenGL state guard (should be identic to
	 *		  global state guard for FBO, different for PBO)
	 *		- returns 0 in case render buffer is not initialized / ready
	 */
	virtual CGLState *p_State();

	/*
	 *	virtual bool CGLRenderBuffer_FBO::b_Status() const
	 *		- returns true in case render buffer, specified by constructor parameters
	 *		  was successfuly created and is ready to use
	 */
	virtual bool b_Status() const;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::CopyTextureData(bool b_release_textures)
	 *		- copies texture data to textures it wasn't possible to directly render to
	 *		- in case b_release_textures is true, all currently bound textures are released
	 *		- returns true on success or false on failure
	 *		- note in this implementation it does not copy no data and always returns true
	 */
	virtual bool CopyTextureData(bool b_release_textures);

	/*
	 *	virtual bool CGLRenderBuffer_FBO::b_DirectRenderToTexture() const
	 *		- returns true in case direct rendering to all textures, being bound is supported,
	 *		  otherwise false
	 *		- note in this implementation it always returns true
	 */
	virtual bool b_DirectRenderToTexture() const;

	/*
	 *	virtual bool CGLRenderBuffer_FBO::Release()
	 *		- releases the render buffer from OpenGL output buffer binding
	 *		- returns true on success or false on failure
	 */
	virtual bool Release();

protected:
	virtual bool Create();
};

/*
 *								=== ~CGLRenderBuffer_FBO ===
 */

// todo - handle rendering to both color and depth textures
// todo - debug fbo-s
// todo - handle rendering to multiple targets

#endif //__RENDER_BUFFER2_INCLUDED
